require("Engine")
-- Hole that can be filled with an object or robot
Hole = Rect:new(0, 0, 0, 0)
function Hole:new(x, y)
	local o = Rect:new(x, y, 24, 24)
	setmetatable(o, self)
	self.__index = self

	o.isHole = true
	o.filled = false

	o.topImg = loadImg("holeblock_top")
	o.bottomImg = loadImg("holeblock_bottom")

	return o
end
function Hole:draw()
	local tr = trans(self)
	love.graphics.draw(self.topImg, tr.x, tr.y)
	if self.contents ~= nil then
		self.contents:draw()
	end
	love.graphics.draw(self.bottomImg, tr.x, tr.y)
end

-- Movable character
Robot = Human:new(0, 0, 0, 0)
function Robot:new(x, y, w, h)
	local o = Human:new(x, y, w, h)
	setmetatable(o, self)
	self.__index = self

	o.animationTable = {moveL = {}, moveR = {}, moveU = {}, moveD = {}}
	o.currAnimation = o.animationTable.moveD
	o.animated = false
	o.drawOffsetX = -(24-o.w)/2
	o.drawOffsetY = -(24-o.h)/2
	o.origDrawOffsetY = o.drawOffsetY
	o.gravity = 0
	o.v = .5
	o.isRobot = true

	o.yOff = 0		-- An offset value for ascending slopes
	o.dYOff = 1 	-- Velocity at which the player falls after descending from a slope
	o.onRamp = false

	o.launched = false
	o.launchTimer = 0
	o.maxLaunchTimer = 20

	return o
end
-- Updates animation based on movement
function Robot:moveAnim()
	if self.x < self.lastX then
		self.animated = true
		self.currAnimation = self.animationTable.moveL
	elseif self.x > self.lastX then
		self.animated = true
		self.currAnimation = self.animationTable.moveR
	end
	if self.y > self.lastY then
		self.animated = true
		self.currAnimation = self.animationTable.moveD
	elseif self.y < self.lastY then
		self.animated = true
		self.currAnimation = self.animationTable.moveU
	end
	-- Disable animation when not moving
	if self.x == self.lastX and self.y == self.lastY then
		self.animated = false
	end
	-- Fall down when not on a slope
	if not self.onRamp then
		if self.yOff < 0 then
			self.drawOffsetY = self.drawOffsetY + self.dYOff
			self.yOff = self.yOff + self.dYOff
		end
	end
	if self.onRamp then
		if self.dx ~= 0 then
			if self.onBlock.dir < 0 then
				if self.dx < 0 then
					self.yOff = self.yOff - self.v/2
					self.drawOffsetY = self.drawOffsetY - self.v/2
				elseif self.dx > 0 and self.yOff < 0 then
					self.yOff = self.yOff + self.v/2
					self.drawOffsetY = self.drawOffsetY + self.v/2
				end
			elseif self.onBlock.dir > 0 then
				if self.dx > 0 then
					self.yOff = self.yOff - self.v/2
					self.drawOffsetY = self.drawOffsetY - self.v/2
				elseif self.dx < 0 and self.yOff < 0 then
					self.yOff = self.yOff + self.v/2
					self.drawOffsetY = self.drawOffsetY + self.v/2
				end
			end
		end
	end
	self.onRamp = false
	self.onBlock = false

	-- Carry out launch actions
	if self.launched then
		self.dx = 0
		self.dy = 0
		if self.launchTimer < 10 then
			self.drawOffsetY = self.drawOffsetY - 10
			self.x = self.x + 48/self.maxLaunchTimer
		elseif self.launchTimer >= 10 and self.launchTimer <= 20 then
			self.drawOffsetY = self.drawOffsetY + 10
			self.x = self.x + 48/self.maxLaunchTimer
		elseif self.launchTimer > 20 then
			self.drawOffsetY = self.origDrawOffsetY
			self.launched = false
		end
		self.launchTimer = self.launchTimer + 1
	end

end
-- Activates launch from a seesaw
function Robot:activateLaunch()
	--self.origDrawOffsetY = self.drawOffsetY
	self.launchTimer = 0
	self.launched = true
end

-- Block that the character slides on
SlipperyTile = Rect:new(0, 0, 0, 0)
function SlipperyTile:new(x, y)
	local o = SlipperyTile:new(x, y, 24, 24)
	setmetatable(o, self)
	self.__index = self

	o.isSlippery = true
	o.img = NamedImg:new("slipperytile")

	return o
end

-- Slope that the player moves up
SlopeTile = Rect:new(0, 0, 0, 0)
function SlopeTile:new(x, y, dir)
	local o = Rect:new(x, y+12, 24, 12)
	setmetatable(o, self)
	self.__index = self

	if dir < 0 then
		self.img = NamedImg:new("slope_l")
	else
		self.img = NamedImg:new("slope_r")
	end

	o.isRamp = true
	o.dir = dir			-- -1 = L, 1 = R; determines how the player moves up a slope

	o.drawOffsetY = -12

	o:findRamps()
	return o
end
-- Sets the ramp that the player is touching
function SlopeTile:action()
	local r = Rect:new(player.x + player.w/2 - 1, player.y + player.h - 1, 2, 2)
	if collideRect(r, self) then
		player.onRamp = true
		player.onBlock = self
	end
end
-- Finds nearby ramps
function SlopeTile:findRamps()
	for i = 1, #ramps do
		if ramps[i] ~= self then
			if self.dir > 0 and approx(ramps[i]:right(), self.x, 2) then
				self.y = self.y + 12
				self.drawOffsetY = self.drawOffsetY - 12
			elseif self.dir < 0 and approx(self:right(), ramps[i].x, 2) then
				self.y = self.y - 12
				self.drawOffsetY = self.drawOffsetY + 12
			end
		end
	end
end

-- Ball that goes back to its original location
Ball = Rect:new(0, 0, 0, 0)
function Ball:new(x, y)
	local o = Rect:new(x, y, 12, 12)
	setmetatable(o, self)
	self.__index = self

	o.gravity = 0
	o.img = NamedImg:new("heavyball")

	o.oldX = x
	o.oldY = y

	return o
end
function Ball:action()
	local nearRobot = false
	-- If it's near a robot, we don't want to reset its position, otherwise, we will
	for i = 1, #entities do
		if entities[i].isRobot and (distToObj(entities[i], self) < 100 or onScreen(self)) then
			nearRobot = true
		end
	end
	if not nearRobot and self.hole == nil then
		self.x = oldX
		self.y = oldY
	end
end

-- Seesaw object
-- Weighted to the right by default; knocking a ball onto it makes the left slam and launches the right side
-- We can either use this to launch a robot to the right over some spikes, or
-- have a robot launch a ball through/onto something
Seesaw = Rect:new(0, 0, 0, 0)
function Seesaw:new(x, y)
	local o = Rect:new(x, y + 12, 36, 12)
	setmetatable(o, self)
	self.__index = self

	o.drawOffsetY = -12
	o.tilt = -1

	o.imgs = {left = NamedImg:new("seesaw_r"), right = NamedImg:new("seesaw_l")}
	o.img = o.imgs.left
	o.pushable = false
	o:action()

	return o
end
function Seesaw:action()
	self.tilt = -1
	for i = 1, #entities do
		local eX = entities[i]:centerX()
		-- Checks for collisions that would launch a right-lying enemy
		if entities[i] ~= self and collideRect(entities[i], self) and eX <= self:centerX() then
			if entities[i].yOff < 0 and entities[i].yOff > -7 then
				self.tilt = 1
			end
		end
	end
	-- Launches a robot to the right
	-- Possible solution: 
	-- Show the robot launching straight up (drawOffsetY)
	-- Once off screen, shift it 2 blocks right of the seesaw's right
	-- Have it rapidly descend (drawOffsetY)
	if self.tilt > 0 then
		for i = 1, #entities do
			if entities[i] ~= self and collideRect(self, entities[i]) and entities[i]:centerX() >= self:centerX() and 
			 entities[i].isRobot then
				entities[i]:activateLaunch()
			end
		end
	end
	-- Determines which direction the seesaw will face
	if self.tilt < 0 then
		self.img = self.imgs.left
	else
		self.img = self.imgs.right
	end
end

-- Displays minitext in a room
TextBlock = Rect:new(0, 0, 0, 0)
function TextBlock:new(x, y, text)
	local o = Rect:new(x, y, 0, 0)
	setmetatable(o, self)
	self.__index = self

	o.img = NamedImg:new("minitextblock")

	o.text = text
	o.addedText = false

	return o
end
-- Display the text once it's on screen
function TextBlock:action()
	if onScreen(self) and not self.addedText then
		self.delete = true
		addMiniText(self.text, {x=self.x, y=self.y})
		self.addedText = true
	end
	if collideRect(self, player) then
		self.x = -1000
		self.y = -1000
	end
end



player = Human:new(50, 50, 16, 14)
player.img = NamedImg:new("robot1_2d")
player.drawOffsetX = -4
player.drawOffsetY = -6
player.gravity = 0
player.v = .5
table.insert(entities, player)
