--require("Entities")
require("Engine")

function loadMap(mapLoc)
	local oldLevel = player.currLevel
	clearMap()
	-- arg[1] gives us the current working directory
	local map = io.open(arg[1].."/Maps/"..mapLoc..".txt", "r")

	player.currLevel = mapLoc

	local currMap = walls

	lightsOn = true
	
	local obj = Rect:new(0, 0, 0, 0)

	magicColorBlocks = {}

	while true do 
		local line = map:read("*line")
		if line == nil then
			break
		end

		local str = splitString(line)
		local x = tonumber(str[2])
		local y = tonumber(str[3])

		local areaMiniText = ""
		local doorConnection = 0		-- Door we connect to
		local missionName = nil			-- Name of a mission to trigger event
		local missionType = 0			-- Type of event that triggers is (1 = announced, 2 = complete, 3 = before)
		local zPos = 0
		local offY = 0
		local offX = 0

		local doSkip = false

		-- Look for possible special cam rects
		if string.find(str[1], "spcam:") then
			table.insert(specialCams, Rect:new(tonumber(str[2]), tonumber(str[3]), tonumber(str[4]), tonumber(str[5])))
		end

		-- Set map's background color
		if string.find(str[1], "bgcolor:") then
			for i = 1, 4 do
				bgColor[i] = tonumber(str[i+1])
			end
		end
		-- Set map's shader (if any)
		if string.find(str[1], "shader:") then
		end

		-- Look for a rain boolean
		if string.find(str[1], "rain:") and string.find(str[2], "true") then
			-- Adds rain
			for i = 1, 4 do
				for j = 1, 4 do
					local fg = Rain:new(160*(i-1), 160*(j-2), 3 * 160)
					table.insert(fgWalls, fg)
				end
			end
		end

		-- Check contents of string beyond 3 for special parameters
		-- Possibilities include links to other doors, special widths/heights, etc
		for i = 4, #str do
			-- Establish a link to another door
			if string.find(str[i], "door=") then
				doorConnection = string.sub(str[i], 6)
			end
			if string.find(str[i], "z=") then
				zPos = tonumber(string.sub(str[i], 3))
			end
			-- Gets text that we want to use as a minitext
			if string.find(str[i], "minitext=") then
				areaMiniText = string.sub(str[i], 10)
				for j = 5, #str do
					areaMiniText = areaMiniText.." "..str[j]
				end
			end
			-- Draw offset values
			if string.find(str[i], "offX=") then
				offX = tonumber(string.sub(str[i], 6))
			elseif 
				string.find(str[i], "offY=") then
				offY = tonumber(string.sub(str[i], 6))
			end
			-- Determine whether an area will have light be default
			if string.find(str[i], "lights=") then
				if string.find(str[i], "off") then
					lightsOn = false
				end
			end
			-- Spawns an entity when a mission is opened and not complete
			if string.find(str[i], "spawnmission=") then
				missionType = 1
				missionName = string.sub(str[i], #"spawnmission=" + 1)
				-- We want to replace underscores with spaces
				local missionDupe = ""
				for i = 1, #missionName do
					if string.sub(missionName, i, i) == "_" then
						missionDupe = missionDupe.." "
					else
						missionDupe = missionDupe..string.sub(missionName, i, i)
					end
				end
				missionName = missionDupe
				-- Check if a mission exists, but is marked as false
				-- If it doesn't exist or is fulfilled, continue
				if not (completedMissions[missionName] == false) then
					doSkip = true
				end
			end
		end

		if string.find(str[1], "background:") then
			currMap = bgWalls
		elseif string.find(str[1], "foreground:") then
			currMap = fgWalls
		end

		-- Sometimes we want to skip loading under certain conditions (e.g., missions)
		if not doSkip then
			if string.find(str[1], "texticon") then
				obj = TextBlock:new(x, y, areaMiniText)
				table.insert(entities, obj)
			elseif string.find(str[1], "hole") and not string.find(str[1], "top") then
				obj = Hole:new(x, y)
				table.insert(currMap, obj)
			elseif string.find(str[1], "narrowcorridor") then
				obj = Rect:new(x, y, 7, 24)
				obj.img = NamedImg:new(str[1])
				obj2 = Rect:new(x+17, y, 7, 24)
				table.insert(currMap, obj2)
				table.insert(currMap, obj)
			elseif string.find(str[1], "slope") then
				if string.find(str[1], "_l") then
					obj = SlopeTile:new(x, y, -1)
				else
					obj = SlopeTile:new(x, y, 1)
				end
				table.insert(ramps, obj)
			elseif string.find(str[1],"ground_") then
				obj = Rect:new(x, y, 0, 0)
				obj.img = NamedImg:new(str[1])
				table.insert(currMap, obj)
			elseif string.find(str[1],"heavyball") then
				obj = Ball:new(x, y)
				table.insert(entities, obj)
			elseif string.find(str[1],"seesaw") then
				obj = Seesaw:new(x, y)
				table.insert(entities, obj)
			elseif string.find(str[1],"slipperytile") then
				obj = Rect:new(x, y, 0, 0)
				obj.img = NamedImg:new(str[1])
				table.insert(currMap, obj)
			elseif string.find(str[1],"bluebeam") then
				obj = Rect:new(x + 8, y, 8, 24)
				obj.drawOffsetX = -8
				obj.img = NamedImg:new(str[1])
				table.insert(walls, obj)


			-- Automatically loads tiles to fit within the dimensions of their image
			-- Operates as a nice fallback for all those wall tiles
			else
				local f = io.open(arg[1].."/Images/"..str[1]..".png", "r")
				if f then
					obj = Rect:new(x, y, 0, 1)
					obj.img = NamedImg:new(str[1])
					obj.w = obj.img.img:getWidth()
					obj.h = obj.img.img:getHeight()
					table.insert(currMap, obj)
				end
			end
			obj.z = zPos
			obj.drawOffsetY = obj.drawOffsetY + offY
			obj.drawOffsetX = obj.drawOffsetX + offX
		end
	end

	-- Find the door the player passed through and set his position to that
	for i = 1, #doors do
		if doors[i].destination == oldLevel then
			player.x = doors[i].x
			player.y = doors[i].y
			break
		end
	end

	-- Get colors for magic color blocks
	local allNotColored = true
	while allNotColored do
		allNotColored = false
		for i = 1, #magicColorBlocks do
			magicColorBlocks[i]:colorize()
			if magicColorBlocks[i].didCol ~= true then
				allNotColored = true
			end
		end
	end

	-- Clear out entities to be removed after certain trigger events have occurred
	local eLen = #entities 
	for i = 1, eLen do
		if i > eLen then
			break
		end
		if entities[i].noSpawn then
			entities[i]:removeSpawn()
			table.remove(entities, i)
			i = i - 1
			eLen = eLen - 1
		end
	end
	-- Clear out entities marked from the previous pass
	local eLen = #entities 
	for i = 1, eLen do
		if i > eLen then
			break
		end
		if entities[i].delete then
			table.remove(entities, i)
			i = i - 1
			eLen = eLen - 1
		end
	end
	for i = 1, #ramps do
		ramps[i]:findRamps()
	end
end

function clearMap()
	walls = {}
	ramps = {}
	--entities = {}
	bgWalls = {}
	fgWalls = {}
	doors = {}
	specialCams = {}
	bullets = {}
	interactiveWalls = {}
	debris = {}

	globalShader = nil
	bgColor = {0, 0, 0, 255}
end



-- For storing/loading data
--[[

TablePersistence is a small code snippet that allows storing and loading of lua variables containing primitive types.
 It is licensed under the MIT license, use it how ever is needed.
 A more detailed description and complete source can be downloaded on 
 http://the-color-black.net/blog/article/LuaTablePersistence.
 A fork has been created on github that included lunatest unit tests: https://github.com/hipe/lua-table-persistence

]]--
local write, writeIndent, writers, refCount;

persistence =
{
	store = function (path, ...)
		local file, e = io.open(path, "w");
		if not file then
			return error(e);
		end
		local n = select("#", ...);
		-- Count references
		local objRefCount = {}; -- Stores reference that will be exported
		for i = 1, n do
			refCount(objRefCount, (select(i,...)));
		end;
		-- Export Objects with more than one ref and assign name
		-- First, create empty tables for each
		local objRefNames = {};
		local objRefIdx = 0;
		file:write("-- Persistent Data\n");
		file:write("local multiRefObjects = {\n");
		for obj, count in pairs(objRefCount) do
			if count > 1 then
				objRefIdx = objRefIdx + 1;
				objRefNames[obj] = objRefIdx;
				file:write("{};"); -- table objRefIdx
			end;
		end;
		file:write("\n} -- multiRefObjects\n");
		-- Then fill them (this requires all empty multiRefObjects to exist)
		for obj, idx in pairs(objRefNames) do
			for k, v in pairs(obj) do
				file:write("multiRefObjects["..idx.."][");
				write(file, k, 0, objRefNames);
				file:write("] = ");
				write(file, v, 0, objRefNames);
				file:write(";\n");
			end;
		end;
		-- Create the remaining objects
		for i = 1, n do
			file:write("local ".."obj"..i.." = ");
			write(file, (select(i,...)), 0, objRefNames);
			file:write("\n");
		end
		-- Return them
		if n > 0 then
			file:write("return obj1");
			for i = 2, n do
				file:write(" ,obj"..i);
			end;
			file:write("\n");
		else
			file:write("return\n");
		end;
		file:close();
	end;

	load = function (path)
		local f, e = loadfile(path);
		if f then
			return f();
		else
			return nil, e;
		end;
	end;
}

-- Private methods

-- write thing (dispatcher)
write = function (file, item, level, objRefNames)
	writers[type(item)](file, item, level, objRefNames);
end;

-- write indent
writeIndent = function (file, level)
	for i = 1, level do
		file:write("\t");
	end;
end;

-- recursively count references
refCount = function (objRefCount, item)
	-- only count reference types (tables)
	if type(item) == "table" then
		-- Increase ref count
		if objRefCount[item] then
			objRefCount[item] = objRefCount[item] + 1;
		else
			objRefCount[item] = 1;
			-- If first encounter, traverse
			for k, v in pairs(item) do
				refCount(objRefCount, k);
				refCount(objRefCount, v);
			end;
		end;
	end;
end;

-- Format items for the purpose of restoring
writers = {
	["nil"] = function (file, item)
			file:write("nil");
		end;
	["number"] = function (file, item)
			file:write(tostring(item));
		end;
	["string"] = function (file, item)
			file:write(string.format("%q", item));
		end;
	["boolean"] = function (file, item)
			if item then
				file:write("true");
			else
				file:write("false");
			end
		end;
	["table"] = function (file, item, level, objRefNames)
			local refIdx = objRefNames[item];
			if refIdx then
				-- Table with multiple references
				file:write("multiRefObjects["..refIdx.."]");
			else
				-- Single use table
				file:write("{\n");
				for k, v in pairs(item) do
					writeIndent(file, level+1);
					file:write("[");
					write(file, k, level+1, objRefNames);
					file:write("] = ");
					write(file, v, level+1, objRefNames);
					file:write(";\n");
				end
				writeIndent(file, level);
				file:write("}");
			end;
		end;
	["function"] = function (file, item)
			-- Does only work for "normal" functions, not those
			-- with upvalues or c functions
			local dInfo = debug.getinfo(item, "uS");
			if dInfo.nups > 0 then
				file:write("nil --[[functions with upvalue not supported]]");
			elseif dInfo.what ~= "Lua" then
				file:write("nil --[[non-lua function not supported]]");
			else
				local r, s = pcall(string.dump,item);
				if r then
					file:write(string.format("loadstring(%q)", s));
				else
					file:write("nil --[[function could not be dumped]]");
				end
			end
		end;
	["thread"] = function (file, item)
			file:write("nil --[[thread]]\n");
		end;
	["userdata"] = function (file, item)
			file:write("nil --[[userdata]]\n");
		end;
}

-- Stores all tables in a save file
function saveAllData()
	local allData = {privlijItems, completedMissions, heldWeapons, states}
	persistence.store("aerannissave.lua", allData)
end
-- Loads all tables
function loadAllData()
	local allData = persistence.load("aerannissave.lua")
	privlijItems = allData[1]
	completedMissions = allData[2]
	heldWeapons = allData[3]
	states = allData[4]
end


