states = {hasCoin = false}
screenWidth = 144
screenHeight = 120
bgColor = {0, 0, 0, 255} 	--The background color of a map

scaleFactor = 3

success = love.graphics.setMode(screenWidth*scaleFactor, screenHeight*scaleFactor, false, true, 0)

require("Engine")
require("Entities")
require("LoadMap")

ctx = love.graphics.newCanvas(screenWidth, screenHeight)
ctx:setFilter("nearest", "nearest")


doors = {}
entities = {}
holeItems = {}
walls = {}


--[[
player = Human:new(50, 50, 16, 14)
player.img = NamedImg:new("robot1_2d")
player.drawOffsetX = -4
player.drawOffsetY = -6
player.gravity = 0
player.v = .5
table.insert(entities, player)
--]]
robot1 = Robot:new(50, 50, 16, 14)
robot1.animationTable = {
	moveD = {NamedImg:new("robot1_1d"), NamedImg:new("robot1_2d"), timer=8},
	moveU = {NamedImg:new("robot1_1u"), NamedImg:new("robot1_2u"), timer=8},
	moveL = {NamedImg:new("robot1_1l"), NamedImg:new("robot1_2l"), timer=8},
	moveR = {NamedImg:new("robot1_1r"), NamedImg:new("robot1_2r"), timer=8}
}
robot1.currAnimation = robot1.animationTable.moveR
robot1:enableAnimation()
robot1.index = 0
table.insert(entities, robot1)

robot2 = Robot:new(100, 50, 10, 14)
robot2.animationTable = {
	moveD = {NamedImg:new("robot2_1d"), NamedImg:new("robot2_2d"), timer=8},
	moveU = {NamedImg:new("robot2_1u"), NamedImg:new("robot2_2u"), timer=8},
	moveL = {NamedImg:new("robot2_1l"), NamedImg:new("robot2_2l"), timer=8},
	moveR = {NamedImg:new("robot2_1r"), NamedImg:new("robot2_2r"), timer=8}
}
robot2.previous = robot1
robot2.doFollow = true
robot2.index = 1
table.insert(entities, robot2)

robot3 = Robot:new(10, 50, 16, 14)
robot3.animationTable = {
	moveD = {NamedImg:new("robot1_1d"), NamedImg:new("robot1_2d"), timer=8},
	moveU = {NamedImg:new("robot1_1u"), NamedImg:new("robot1_2u"), timer=8},
	moveL = {NamedImg:new("robot1_1l"), NamedImg:new("robot1_2l"), timer=8},
	moveR = {NamedImg:new("robot1_1r"), NamedImg:new("robot1_2r"), timer=8}
}
robot3.previous = robot2
robot3.doFollow = true
robot3.index = 2
table.insert(entities, robot3)

player = robot1
robot1.next = robot2
robot2.next = robot3
robot3.next = nil


-- Fills non-transparent pixels with black
blackShader = love.graphics.newPixelEffect[[
	extern number time;
	extern number scale;
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords)
    {
		vec4 pix = texture2D(texture, vec2(texture_coords.x, texture_coords.y));
		float f = mod(screen_coords.y, scale);
		if (f < 1.3 / time){
			pix.b = pix.b * .6;
			pix.r = 0;
			pix.g = pix.g * .8 * time;
		}
		else {
			pix.b = pix.b * time;
		}

		f = mod(screen_coords.x, scale);
		if (f < 1){
			pix.b = pix.b * .7;
		}

		return pix;
	}
]]

-- Keeps track of steps for following others
followSteps = {}
-- Number of steps behind the object it's following
totalSteps = 35

-- Used to sort by y position
function compare(item1, item2)
	return item1.y < item2.y
end

-- Swaps out the player with the last robot
function swapLeader(leader)
	local old = player
	old.dx = 0
	old.dy = 0
	local final = 0
	-- Get the last robot in the queue
	local curr = old
	while curr.next ~= nil do
		curr = curr.next
	end
	curr.previous.next = nil
	curr.next = player
	player.previous = curr
	player = curr

	-- Also dismiss other robots
	dismissRobots()
end


-- Remove a robot from the follow list, and set what follows it to follow the previous
function removeRobot(robot)
	robot.doFollow = false
	robot.abandoned = true
	-- Remove robot from linked list
	if robot.previous then
		robot.previous.next = robot.next
	end
	if robot.next then
		robot.next.previous = robot.previous
	end
	robot.next = nil
	robot.previous = nil
	--[[
	for i = 1, #entities do
		if entities[i].previous and entities[i].previous == robot then
			entities[i].index = robot.index
			entities[i].previous = robot.previous
		end
	end
	]]
end

-- Tells all robots to stop following
function dismissRobots()
	for i = 1, #entities do
		entities[i].doFollow = false
	end
end

-- Change control to another robot when dismissed to allow them to explore alone
function cycleControl()

end

-- Pulls trailing entities, like how we did octopus arm movement in Subbania
function pullFollow(e)
	local pv = player.v
	if e.previous then
		local p = e.previous
		if e.x > p:right() then
			e.x = p:right()
			-- Nudge toward center
			if e.y < p.y then
				e.y = e.y + pv
			elseif e.y > p.y then
				e.y = e.y - pv
			end
		elseif e:right() < p.x then
			e.x = p.x - p.w
			-- Nudge toward center
			if e.y < p.y then
				e.y = e.y + pv
			elseif e.y > p.y then
				e.y = e.y - pv
			end
		end
		if e.y > p:bottom() then
			e.y = p:bottom()
			-- Nudge toward center
			if e.x < p.x then
				e.x = e.x + pv
			elseif e.x > p.x then
				e.x = e.x - pv
			end
		elseif e:bottom() < p.y then
			e.y = p.y - e.h
			-- Nudge toward center
			if e.x < p.x then
				e.x = e.x + pv
			elseif e.x > p.x then
				e.x = e.x - pv
			end
		end
	end
end


-- Push a robot into a "block" (i.e., hole) to remove it
function removeHole(robot, block)
	if robot.doFollow == false and block.isHole == true then
		if collideRect(robot, block) and (robot.pushDx ~= 0 or robot.pushDy ~= 0) then
			block.filled = true
			robot.hole = block		-- Set the hole we want it to move to
			block.contents = robot
			removeRobot(robot)
		end
	end
end

-- Move the robot towards the hole, then have it sink it
function moveToHole(obj, index)
	if obj.isRobot then
		-- Can't sacrifice the skinny robot--he won't fit the hole
		if obj.index == 1 then
			return 0
		end
		-- Otherwise, sacrificing them is good
		removeRobot(obj)
	end
	obj.hole.w = 0
	obj.hole.h = 0
	obj.dx = 0
	obj.dy = 0
	obj.immobile = true
	obj.hole.filled = true
	local nearX = true
	local nearY = true
	local slideSpeed = .3
	-- Move towards hole if not near it
	local xOff = 4
	if not approx(obj.x - xOff, obj.hole.x, 1) then
		local x1 = obj.x - xOff - obj.hole.x
		if x1 == 0 then
			x1 = 1
		end
		obj.x = obj.x - (obj.x - xOff - obj.hole.x)/math.abs(x1) * slideSpeed
		nearX = false
	end
	if not approx(obj.y, obj.hole.y, 1) then
		local y1 = obj.y - obj.hole.y
		if y1 == 0 then
			y1 = 1
		end
		obj.y = obj.y - (obj.y - obj.hole.y)/math.abs(y1) * slideSpeed
		nearY = false
	end

	-- If near the hole but not fallen into it, fall down
	if nearX and nearY then
		-- We'll now want to draw as being inside the hole
		obj.hole.contents = obj
		obj.invisible = true

		if obj.drawOffsetY < 0 then
			obj.drawOffsetY = obj.drawOffsetY + slideSpeed
		-- If down completely, move it to another list that's below the hole tile
		else
			obj.putIntoHole = true
			--table.remove(entities, index)
			-- Put the object inside the hole and keep it there
			--table.insert(holeItems, obj)
		end
	end
end

--[[

This is a game about a small group of robots. They go on a journey to find the last 
engineer in order to get themselves repaired.

These robots are the last of their kind. All other life on earth is dead. The engineer is 
only known in legends, and there's a chance they will all die.

Create a Zelda-like map system. Only allow the player to move onto spaces with a tile.

]]--

--[[
	SCENE IDEAS:
		1: There's an unpassable shallow hole. You have two choices: push one robot in the hole to cross,
			or have a system like 2D Zeldas where you move right about 5 repeated screens and 
			discover a path across.
			Have a rock on the other side that you can push into the hole to let the other robots cross.
]]--

-- You sacrificed your fellow robot for some useless item
function greedyEnding()

end
-- You avoided treasure and had 2+ robots
function betterEnding()

end

loadMap("map1")
function love.load()
	love.graphics.setPixelEffect(blackShader)
end
table.insert(walls, Hole:new(10, 10))
function keyPress(dt)
	if love.keyboard.isDown("right") then
		--player.dx = 2
		player.totalMoveTimer = player.totalMoveTimer + 1
		player.facing = 1
		player.direction = 1
		rightUp = false
	else
		rightUp = true
	end
	
	if love.keyboard.isDown("left") then
		--player.dx = -2
		player.totalMoveTimer = player.totalMoveTimer + 1
		player.facing = -1
		player.direction = -1
		leftUp = false
	else
		leftUp = true
	end
	
	if not love.keyboard.isDown("right") and not love.keyboard.isDown("left") then
		--player.dx = 0
		-- Reset timer for total time walked/run
		if player.totalMoveTimer > 0 then
			player.totalMoveTimer = player.totalMoveTimer - 1 
		end
		player.direction = 0
	end
	
	-- Triggers events and interact with characters
	if love.keyboard.isDown("up") and upUp then
		player.dz = -.05
		--[[
		for i = 1, #entities do
			if entities[i].interactive and collideRect(player, entities[i]) then
				entities[i]:interact()
			end
		end
		
		for i = 1, #doors do
			-- Enter a door and move to the next area
			if doors[i].interactive and collideRect(player, doors[i]) then
				doors[i]:interact()
				break
			end
		end]]
		upUp = false

	end
	if not love.keyboard.isDown("up") then
		upUp = true
	end

	-- Jumping for player (or text advance in those situations)
	advanceText = false
	if love.keyboard.isDown("x") then
		player:jump()
		if xUp then
			advanceText = true
			-- Swaps up which character is in front
			swapLeader(player)
		end
		xUp = false
	end
	
	if not love.keyboard.isDown("x") then
		if player.dy < 0 then
			player.dy = -player.dy/4
		end
		xUp = true
	end
	
	-- Activate crouch
	if love.keyboard.isDown("down") and downUp then
		player.dz = .05
		downUp = false
	end
	
	if not love.keyboard.isDown("down") then
		downUp = true
	end
	if downUp and upUp then
		player.dz = 0
	end
	
	-- Melee attacks for player
	if love.keyboard.isDown("z") then
		player.meleeTriggered = true
		player.attacking = true
		zUp = false
	end
	if not love.keyboard.isDown("z") then
		zUp = true
		player.attacking = false
	end

	-- Shooting for player
	if love.keyboard.isDown("c") and cUp then
		cUp = false
		dismissRobots()
	end
	if not love.keyboard.isDown("c") and not cUp then
		cUp = true
	end

	-- Activates pause menu
	if love.keyboard.isDown("return") and enterUp then
		enterUp = false
		states.paused = true
	end
	if not love.keyboard.isDown("return") then
		enterUp = true
	end
end

-- Game logic
function love.update(dt)
	keyPress()

	-- Move player
	if not downUp then
		player.dy = player.v
	elseif not upUp then
		player.dy = -player.v
	elseif upUp and downUp then
		player.dy = 0
	end
	if not rightUp then
		player.dx = player.v
	elseif not leftUp then
		player.dx = -player.v
	elseif leftUp and rightUp then
		player.dx = 0
	end
	for i = 1, #bgWalls do
		bgWalls[i]:draw()
	end
	for i = 1, #walls do
		walls[i]:action()
	end

	for i = 1, #ramps do
		ramps[i]:action()
	end
	for i = 1, #entities do
		-- Store the old position for updating animations
		entities[i].lastX = entities[i].x
		entities[i].lastY = entities[i].y
		if entities[i] ~= player then
			-- Move to the hole
			if entities[i].colWith ~= nil and entities[i].colWith.isHole then
				entities[i].hole = entities[i].colWith
			end
			if entities[i].hole ~= nil then
				moveToHole(entities[i], i)
			end

			entities[i]:action()

			-- Allows entities to push/be pushed
			if not (entities[i].pushable == false) then
				-- Speed combines normal speed + push speed
				entities[i].dx = entities[i].dx + entities[i].pushDx
				entities[i].dy = entities[i].dy + entities[i].pushDy
				
				entities[i]:move()

				-- We want to reset push velocity
				entities[i].dx = entities[i].dx - entities[i].pushDx
				entities[i].dy = entities[i].dy - entities[i].pushDy

				entities[i].pushDx = 0
				entities[i].pushDy = 0
			end
		end
	end

	local didReduceDy = false
	local didReduceDx = false
	-- Allow the player to push entities
	for i = 1, #entities do
		if collideRect(entities[i], player) and entities[i] ~= player and not entities[i].immobile then
			-- We test for a push within a reasonable range of the velocity
			local r = (math.abs(player.dx) + math.abs(player.dy))*2
			-- Push object horizontally
			if player.dx ~= 0 then
				if (player.dx < 0 and player.x > entities[i]:right() - r) 
				 or (player.dx > 0 and player:right() < entities[i].x + r) then
					if not didReduceDx then
						didReduceDx = false
						player.dx = player.dx / 2
						entities[i].pushDx = player.dx
					end
				end
			end
			-- Push object vertically
			if player.dy ~= 0 then
				if (player.dy < 0 and player.y > entities[i]:bottom() - r) or (player.dy > 0 and player:bottom() < entities[i].y + r) then
				 	if not didReduceDy then
						didReduceDy = false
						player.dy = player.dy / 2
						entities[i].pushDy = player.dy
					end
				end
			end
		end
		-- Entities following an object take their position from n movements ago
		-- Alternatively, we could have something like rect dragging, like we did with
		-- the octopus arms in Subbania
		----[[
		local steps = #followSteps
		if entities[i] ~= player and entities[i].doFollow and #followSteps > totalSteps*entities[i].index then
			local oldX = entities[i].x
			local oldY = entities[i].y
			entities[i].x = followSteps[steps - totalSteps*entities[i].index][1]
			entities[i].y = followSteps[steps - totalSteps*entities[i].index][2]
			
			-- Get direction faced
			if entities[i].x < oldX then
				entities[i].xFace = -1
			elseif entities[i].x > oldX then
				entities[i].xFace = 1
			end
			if entities[i].y < oldY then
				entities[i].yFace = -1
			elseif entities[i].y > oldY then
				entities[i].yFace = 1
			end

			-- Check to see if we collided with a wall
			for j = 1, #walls do
				if collideRect(entities[i], walls[j]) then

				end

			end
		end
		if entities[i] ~= player then
			entities[i]:moveAnim()
		end
		if entities[i].currAnimation and entities[i].animated then
			entities[i]:animate()
		end
		--]]
		--pullFollow(entities[i])
	end

	player:action()
	player:move()
	player:moveAnim()

	-- Insert former positions of the player for following behavior
	if player.dx ~= 0 or player.dy ~= 0 then
		table.insert(followSteps, {player.x, player.y})
	end
	-- Remove old steps
	if #followSteps > #entities * totalSteps then
		table.remove(followSteps, 1)
	end

	-- Pan the camera to the next area
	if (player.x + player.w/2) < camRect.x then
		camRect.x = camRect.x - screenWidth
	elseif (player.x + player.w/2) > camRect:right() then
		camRect.x = camRect.x + screenWidth
	end
	if (player.y + player.h/2) < camRect.y then
		camRect.y = camRect.y - screenHeight
	elseif (player.y + player.h/2) > camRect:bottom() then
		camRect.y = camRect.y + screenHeight
	end
end






-- Used to give the game a nice screen flicker
flickerV = .8
-- Draw code
function love.draw()
	flickerV = flickerV - .1
	if flickerV < .85 then
		flickerV = 1.25
	end
	blackShader:send("time", flickerV)
	blackShader:send("scale", scaleFactor)

	resetEffects()
	love.graphics.setCanvas(ctx)
	ctx:clear(bgColor[1], bgColor[2], bgColor[3], bgColor[4])

	for i = 1, #bgWalls do
		if nearScreen(bgWalls[i], 24) then
			bgWalls[i]:draw()
		end
	end
	for i = 1, #walls do
		walls[i]:draw()
	end
	for i = 1, #ramps do
		ramps[i]:draw()
	end

	-- Sorts entities before drawing
	table.sort(entities, compare)
	for i = 1, #entities do
		-- Sometimes entities become invisible
		if not entities[i].invisible then
			entities[i]:draw()
		end
	end

	-- MiniTexts
	local miniTextLen = #miniTexts
	for i = 1, miniTextLen do
		if i > miniTextLen then
			break
		end
		miniTexts[i]:draw()
		if miniTexts[i].delete then
			table.remove(miniTexts, i)
			i = i - 1
			miniTextLen = miniTextLen - 1
		end
	end

	love.graphics.setCanvas() --This sets the target back to the screen
	love.graphics.setPixelEffect(blackShader)
	love.graphics.draw(ctx, 0, 0, 0, scaleFactor, scaleFactor)
end